# B0003

As commands are executed asynchronously, it is possible to issue a command on an entity that will no longer exist at the time of the command execution.

Erroneous code example:

```rust,should_panic
use bevy::prelude::*;

fn main() {
    App::new()
        .add_plugins(DefaultPlugins)
        .add_startup_system(setup)
        .add_system(despawning)
        .add_system(use_entity.after(despawning))
        .run();
}

struct MyEntity(Entity);

#[derive(Component)]
struct Hello;

fn setup(mut commands: Commands) {
    let entity = commands.spawn().id();
    commands.insert_resource(MyEntity(entity));
}

fn despawning(mut commands: Commands, entity: Option<Res<MyEntity>>) {
    if let Some(my_entity) = entity {
        commands.entity(my_entity.0).despawn();
        commands.remove_resource::<MyEntity>();
    }
}

fn use_entity(mut commands: Commands, entity: Option<Res<MyEntity>>) {
    if let Some(my_entity) = entity {
        commands.entity(my_entity.0).insert(Hello);
    }
}
```

This will panic, as system `use_entity` is executed after system `despawning`. Without the system ordering specified here, the ordering would be random and this code would panic half the time.

The default panic message is telling you the entity id (`0v0`) and the command that failed (adding a component `Hello`):

```text
thread 'main' panicked at 'error[B0003]: Could not add a component (of type `use_entity_after_despawn::Hello`) to entity 0v0 because it doesn't exist in this World.', /bevy/crates/bevy_ecs/src/system/commands/mod.rs:752:13
```

But you don't know which system tried to add a component, and which system despawned the entity.

To get the system that created the command that panics, you can enable the `trace` feature of Bevy. This will add a panic handler that will print more informations:

```text
   0: bevy_ecs::schedule::stage::system_commands
           with name="use_entity_after_despawn::use_entity"
             at crates/bevy_ecs/src/schedule/stage.rs:880
   1: bevy_ecs::schedule::stage
           with name=Update
             at crates/bevy_ecs/src/schedule/mod.rs:337
   2: bevy_app::app::frame
             at crates/bevy_app/src/app.rs:113
   3: bevy_app::app::bevy_app
             at crates/bevy_app/src/app.rs:126
thread 'main' panicked at 'error[B0003]: Could not add a component (of type `use_entity_after_despawn::Hello`) to entity 0v0 because it doesn't exist in this World.', /bevy/crates/bevy_ecs/src/system/commands/mod.rs:752:13
```

From the first two lines, you now know that it panics while executing a command from the system `use_entity`.

To get the system that created the despawn command, you can enable DEBUG logs for crate `bevy_ecs`, for example by setting the environment variable `RUST_LOG=bevy_ecs=debug`. This will log:

```text
DEBUG stage{name=Update}:system_commands{name="use_entity_after_despawn::despawning"}: bevy_ecs::world: Despawning entity 0v0
thread 'main' panicked at 'error[B0003]: Could not add a component (of type `use_entity_after_despawn::Hello`) to entity 0v0 because it doesn't exist in this World.', /bevy/crates/bevy_ecs/src/system/commands/mod.rs:752:13
```

From the first line, you know the entity `0v0` was despawned when executing a command from system `despawning`. In a real case, you could have many log lines, you will need to search for the exact entity from the panic message.

Combining those two, you should get enough informations to understand why this panic is happening and how to fix it:

```text
DEBUG stage{name=Update}:system_commands{name="use_entity_after_despawn::despawning"}: bevy_ecs::world: Despawning entity 0v0
   0: bevy_ecs::schedule::stage::system_commands
           with name="use_entity_after_despawn::use_entity"
             at crates/bevy_ecs/src/schedule/stage.rs:880
   1: bevy_ecs::schedule::stage
           with name=Update
             at crates/bevy_ecs/src/schedule/mod.rs:337
thread 'main' panicked at 'error[B0003]: Could not add a component (of type `use_entity_after_despawn::Hello`) to entity 0v0 because it doesn't exist in this World.', /bevy/crates/bevy_ecs/src/system/commands/mod.rs:752:13
```
